require "option_parser"
require "ipc"
require "socket"
require "./colors"
require "./ws"

require "socket"
require "http"
require "http/server"
require "base64"
require "digest"
require "./utils"

class IPC::Connection
	def initialize(fd : Int32)
		external_connection = LibIPC::Connection.new
		external_connection.fd = fd
		initialize(external_connection)
	end
end

class IPC::Message
	def to_buffer
		to_message @type, String.new(@payload)
	end
end


class WrappedTCPFileDescriptor < TCPSocket
	# do not close the connection when garbage collected!!
	def finalize
		# puts "WrappedTCPFileDescriptor garbage collection!!"
		# super
	end
end

service_name = "websocket"
port_to_listen = 1234

OptionParser.parse! do |parser|
	parser.on "-p port", "--port port", "Port to listen on." do |port|
		port_to_listen = port.to_u16
	end

	parser.on "-s service-name", "--service-name service-name", "Service name." do |name|
		service_name = name
	end

	parser.on "-h", "--help", "Show this help" do
		puts parser
		exit 0
	end
end


class InstanceStorage
	property service : IPC::SwitchingService
	property is_client : Hash(Int32,Bool)
	property switchtable : Hash(Int32, Int32)
	property fdtotcpsocket : Hash(Int32, TCPSocket)
	property fdtows : Hash(Int32, WebSocket)
	property fdtoipcconnection : Hash(Int32, IPC::Connection)

	def initialize (@service : IPC::SwitchingService)
		# fdlist_client = [] of TCPSocket
		@is_client = Hash(Int32,Bool).new
		@fdtotcpsocket = Hash(Int32, TCPSocket).new
		@fdtows = Hash(Int32, WebSocket).new
		@switchtable = Hash(Int32, Int32).new
		@fdtoipcconnection = Hash(Int32, IPC::Connection).new
	end

	def remove_fd (fdclient : Int32)
		# 1. closing both the client and the service
		fdservice = @switchtable[fdclient]?
		tcpfdc = @fdtotcpsocket[fdclient]

		# 2. closing the TCP connections
		tcpfdc.close unless tcpfdc.closed?

		# 3. removing the client and the service fds from the loop check
		@service.remove_fd (fdclient)

		# 5. removing both the client and the service from the switchtable
		@switchtable = @switchtable.select do |fdc, fds|
			fdc != fdclient && fds != fdclient
		end

		# 6. removing the client and the service from is_client
		@is_client = @is_client.select do |fd,v| fd != fdclient end

		unless fdservice.nil?
			service = @fdtoipcconnection[fdservice]
			service.close
			@service.remove_fd (fdservice)
			@fdtoipcconnection = @fdtoipcconnection.select do |k, v|
				k != fdservice
			end

			@is_client = @is_client.select do |fd,v| fd != fdservice end
		end
	end
end

server = TCPServer.new("localhost", port_to_listen)
service = IPC::SwitchingService.new service_name
service << server.fd
context = InstanceStorage.new service



def websocket_client_connection(client, context : InstanceStorage)
	request = HTTP::Request.from_io client
	# pp! request

	if request.nil?
		raise "#REQUEST IS NIL"
	end

	if request.is_a? HTTP::Request::BadRequest
		raise "BAD REQUEST DAZE~"
	end

	# FIXME: check they actually wanted to upgrade to websocket
	key = request.headers["Sec-WebSocket-Key"]
	response_key = Digest::SHA1.base64digest key + "258EAFA5-E914-47DA-95CA-C5AB0DC85B11"
	# puts response_key

	# HTTP inside bru
	headers_header = "HTTP/1.1 101 Switching Protocols"
	headers = HTTP::Headers {
		"Upgrade" => "websocket",
		"Connection" => "Upgrade",
		"Sec-WebSocket-Accept" => response_key
	}

	headers = headers.map { |key, value| "#{key}: #{value[0]}\r\n" }.join

	# requested service, fd
	req_service = request.path.lchop
	if req_service.empty?
		client.close
		puts "should send a PATH"
		return
	end

	websocket_connection_procedure req_service, client.fd, context

	# puts "#{headers_header}\n#{headers.to_s}\r\n"
	client.send "#{headers_header}\n#{headers.to_s}\r\n"

	wsclient = WebSocket.new client

	# the client is still not connected to a service
	context.is_client[client.fd] = true

	# listen to the client's file descriptor
	context.service << client.fd
	# puts "#{CBLUE}new client: #{client.fd}#{CRESET}"

	# registering the client into storing structures to avoid being garbage collected
	context.fdtotcpsocket[client.fd] = client
	context.fdtows[client.fd] = wsclient
end


def closing_client (fdclient : Int, context : InstanceStorage)
	puts "#{CBLUE}closing the client:#{CRESET} #{fdclient}"
	context.remove_fd fdclient
end

# first message from the client: requested service name
# 1. connection to the service
# 2. listening on the service fd
# 3. bounding both file descriptors (through switchtable hash)
# 4. indicating that the client is connected (in is_client)
def websocket_connection_procedure (requested_service : String, clientfd : Int32, context : InstanceStorage)
	begin
		# 2. establishing a connection to the service
		newservice = IPC::Connection.new requested_service
		context.fdtoipcconnection[newservice.fd] = newservice

		# 3. listening on the client fd and the service fd
		context.service << newservice.fd

		# cannot perform automatic switching due to websockets headers
		# future version of the libipc lib should include some workaround, probably
		# service.switch.add fdclient, newservice.fd

		# 4. bounding the client and the service fd
		context.switchtable[clientfd] = newservice.fd
		context.switchtable[newservice.fd] = clientfd

		# 5. the client is then connected, send it a message
		context.is_client[clientfd] = true
		context.is_client[newservice.fd] = false
	rescue e
		puts "#{CRED}Exception during connection to the service:#{CRESET} #{e}"
		closing_client clientfd, context
	end
end

def websocket_switching_procedure (activefd : Int, context : InstanceStorage)
	begin
		if context.is_client[activefd]
			# The client is a WebSocket on top of a TCP connection
			client = context.fdtotcpsocket[activefd]
			wsclient = context.fdtows[activefd]
			begin
				message = wsclient.read
			rescue e
				puts "#{CRED}Exception (receiving a message)#{CRESET} #{e}"
				closing_client activefd, context
				return
			end

			if wsclient.closed?
				# puts "#{CBLUE}client is closed#{CRESET}"
				closing_client client.fd, context
				return
			end

			if message.nil?
				puts "#{CRED}message is nil#{CRESET}"
				closing_client client.fd, context
				return
			end

			# puts "switching: client to service"
			# print_hexa(String.new(message), "#{CBLUE}Received message hexa#{CRESET}")
			# client => service
			fdservice = context.switchtable[activefd]

			# XXX: this is not a TCP fd, but since behind the scene this is compatible, I'm hacking a bit
			serv = WrappedTCPFileDescriptor.new(fd: fdservice, family: Socket::Family::INET)
			#serv = context.fdtoipcconnection[fdservice]
			serv.send message

		else
			# puts "switching: service to client"
			# service => client

			fdclient = context.switchtable[activefd]
			wsclient = context.fdtows[fdclient]

			serv = context.fdtoipcconnection[activefd]
			message = serv.read
			# puts "received message from service: #{message.to_s}"
			buf = message.to_buffer
			# print_hexa String.new(buf), "\033[31m Message to send to the client \033[00m "
			wsclient.send buf
		end
	rescue e
		puts "#{CRED}Exception during message transfer:#{CRESET} #{e}"
		if context.is_client[activefd]
			closing_client activefd, context
		else
			clientfd = context.switchtable[activefd]
			closing_client clientfd, context
		end
	end

end

service.loop do |event|
	case event
	when IPC::Event::Connection
		puts "#{CBLUE}IPC::Event::Connection: #{event.connection.fd}#{CRESET}"
	when IPC::Event::Disconnection
		puts "#{CBLUE}IPC::Event::Disconnection: #{event.connection.fd}#{CRESET}"
	when IPC::Event::ExtraSocket
		puts "#{CBLUE}IPC::Event::ExtraSocket#{CRESET}"

		# 1. accept new websocket clients
		if server.fd == event.connection.fd
			client = server.accept
			begin
				websocket_client_connection client, context
				puts "#{CBLUE}new client:#{CRESET} #{client.fd}"
			rescue e
				puts "Exception: #{CRED}#{e}#{CRESET}"
				client.close
			end
			next
		end

		# 2. active fd != server fd
		activefd = event.connection.fd
		websocket_switching_procedure activefd, context

	when IPC::Event::Switch
		puts "\033[36mIPC::Event::Switch#{CRESET}: from fd #{event.connection.fd}"

		raise "Not implemented."

	# IPC::Event::Message has to be the last entry
	# because ExtraSocket and Switch inherit from Message class
	when IPC::Event::Message
		puts "#{CBLUE}IPC::Event::Message#{CRESET}: #{event.connection.fd}"

		raise "Not implemented."
	end
end
